"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

import re
import urllib.error
import urllib.parse
import urllib.request
from logging import getLogger
from typing import Union

from boilerpy3 import filters, parser
from boilerpy3.document import TextDocument
from boilerpy3.exceptions import HTMLExtractionError
from boilerpy3.filters import BoilerpipeFilter
from boilerpy3.marker import HTMLBoilerpipeMarker

logger = getLogger('boilerpy3')


class Extractor:
    """
    The base class of Extractors. Also provides some helper methods to quickly retrieve the text that remained after
    processing.
    """
    
    SCRIPT_REGEX = re.compile(r'<(?:script|SCRIPT)[^>]*>.*?</(?:script|SCRIPT)>', re.DOTALL)
    
    def __init__(self, filtr: BoilerpipeFilter, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param filtr: filter
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        self.filter = filtr
        self.raise_on_failure = raise_on_failure
    
    def get_content(self, text: str) -> str:
        return self.get_doc(text).content
    
    def get_content_from_url(self, url: str) -> str:
        return self.get_doc_from_url(url).content
    
    def get_content_from_file(self, filename) -> str:
        return self.get_doc_from_file(filename).content
    
    def get_doc_from_file(self, filename) -> TextDocument:
        return self.get_doc(self.read_from_file(filename))
    
    def get_doc_from_url(self, url) -> TextDocument:
        return self.get_doc(self.read_from_url(url))
    
    def get_doc(self, text) -> TextDocument:
        doc = self.parse_doc(text)
        self.filter.process(doc)
        return doc
    
    def get_marked_html(self, text) -> str:
        doc = self.get_doc(text)
        m = HTMLBoilerpipeMarker(raise_on_failure=self.raise_on_failure)
        return m.process(doc, text)
    
    def read_from_file(self, filename: str) -> str:
        with open(filename) as text_file:
            return text_file.read()
    
    def read_from_url(self, url: str) -> str:
        with urllib.request.urlopen(url) as url_obj:
            text = url_obj.read()
            encoding = self.get_url_encoding(url_obj)
        
        try:
            text = text.decode(encoding)
        except UnicodeDecodeError:
            pass
        return text
    
    def get_url_encoding(self, f) -> str:
        try:
            return f.headers['content-type'].split('charset=')[1].split(';')[0]
        except:
            return 'utf8'
    
    def parse_doc(self, input_str: str) -> Union[TextDocument, None]:
        bp_parser = parser.BoilerpipeHTMLParser(raise_on_failure=self.raise_on_failure)
        try:
            bp_parser.feed(input_str)
        except:
            # in case of error, try again, first removing script tag content
            bp_parser = parser.BoilerpipeHTMLParser(raise_on_failure=self.raise_on_failure)
            input_str = self.SCRIPT_REGEX.sub('<script></script>', input_str)
            try:
                bp_parser.feed(input_str)
            except Exception as ex:
                logger.exception('Error parsing HTML')
                if self.raise_on_failure:
                    raise HTMLExtractionError from ex
                else:
                    return TextDocument([])
        doc = bp_parser.to_text_document()
        return doc


class DefaultExtractor(Extractor):
    """
    Usually worse than ArticleExtractor, but simpler/no heuristics. A quite generic full-text extractor.
    """
    
    _filter_chain = filters.FilterChain([
        filters.SimpleBlockFusionProcessor(),
        filters.BlockProximityFusion(1, False, False),
        filters.DensityRulesClassifier()
    ])
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter_chain, raise_on_failure)


class ArticleExtractor(Extractor):
    """
    A full-text extractor which is tuned towards news articles. In this scenario it achieves higher accuracy than
    DefaultExtractor. Works very well for most types of Article-like HTML.
    """
    
    _filter_chain = filters.FilterChain([
        filters.TerminatingBlocksFinder(),
        filters.DocumentTitleMatchClassifier(None, True),
        filters.NumWordsRulesClassifier(),
        filters.IgnoreBlocksAfterContentFilter(),
        filters.BlockProximityFusion(1, False, False),
        filters.BoilerplateBlockFilter(),
        filters.BlockProximityFusion(1, True, False),
        filters.KeepLargestBlockFilter(),
        filters.ExpandTitleToContentFilter()
    ])
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter_chain, raise_on_failure)


class LargestContentExtractor(Extractor):
    """
    A full-text extractor which extracts the largest text component of a page. For news articles, it may perform better
    than the DefaultExtractor, but usually worse than ArticleExtractor. Like DefaultExtractor, but keeps the largest
    text block only.
    """
    
    _filter_chain = filters.FilterChain([
        filters.NumWordsRulesClassifier(),
        filters.BlockProximityFusion(1, False, False),
        filters.KeepLargestBlockFilter()
    ])
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter_chain, raise_on_failure)


class CanolaExtractor(Extractor):
    """
    Trained on krdwrd Canola (different definition of "boilerplate"). You may give it a try.
    """
    
    _filter = filters.CanolaFilter()
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter, raise_on_failure)


class KeepEverythingExtractor(Extractor):
    """
    Marks everything as content. Dummy Extractor; should return the input text. Use this to double-check that your
    problem is within a particular BoilerpipeExtractor, or somewhere else.
    """
    
    _filter = filters.MarkEverythingContentFilter()
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter, raise_on_failure)


class NumWordsRulesExtractor(Extractor):
    """
    A quite generic full-text extractor solely based upon the number of words per block (the current, the previous and
    the next block).
    """
    
    _filter = filters.NumWordsRulesClassifier()
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter, raise_on_failure)


class ArticleSentencesExtractor(Extractor):
    """
    A full-text extractor which is tuned towards extracting sentences from news articles.
    """
    
    _filter_chain = filters.FilterChain([
        ArticleExtractor._filter_chain,
        filters.SplitParagraphBlocksFilter(),
        filters.MinClauseWordsFilter()
    ])
    
    def __init__(self, raise_on_failure: bool = True) -> None:
        """
        Initialize extractor
        
        :param raise_on_failure: whether or not to raise an exception if a text extraction failure is encountered.
        """
        
        super().__init__(self._filter_chain, raise_on_failure)


class KeepEverythingWithMinKWordsFilter(filters.FilterChain):
    """
    A full-text extractor which extracts the largest text component of a page. For news articles, it may perform better
    than the DefaultExtractor, but usually worse than ArticleExtractor.
    """
    
    def __init__(self, k_min: int) -> None:
        # Note: variable was not used initially, seems it should be passed to super() call
        filter_arr = [
            filters.SimpleBlockFusionProcessor(),
            filters.MarkEverythingContentFilter(),
            filters.MinWordsFilter(k_min)
        ]
        super().__init__(filter_arr)



"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

import copy
import sys
from typing import List, Set, Tuple, Union


class DefaultLabels:
    """
    Some pre-defined labels which can be used in conjunction with TextBlock
    """
    
    TITLE = "de.l3s.boilerpipe/TITLE"
    ARTICLE_METADATA = "de.l3s.boilerpipe/ARTICLE_METADATA"
    INDICATES_END_OF_TEXT = "de.l3s.boilerpipe/INDICATES_END_OF_TEXT"
    MIGHT_BE_CONTENT = "de.l3s.boilerpipe/MIGHT_BE_CONTENT"
    STRICTLY_NOT_CONTENT = "de.l3s.boilerpipe/STRICTLY_NOT_CONTENT"
    HR = "de.l3s.boilerpipe/HR"
    MARKUP_PREFIX = "<"


class TextBlock:
    """
    Describes a block of text. A block can be an "atomic" text element (i.e., a sequence of text that is not interrupted
    by any HTML markup) or a compound of such atomic elements.
    """
    
    EMPTY_START: 'TextBlock'
    EMPTY_END: 'TextBlock'
    
    def __init__(self, text: str, contained_text_elements=None, num_words: int = 0, num_words_in_anchor_text: int = 0,
                 num_words_in_wrapped_lines: int = 0, num_wrapped_lines: int = 0, offset_blocks: int = 0) -> None:
        self.is_content = False
        self.labels = set()
        self.num_full_text_words = 0
        self.tag_level = 0
        
        self.text = text
        if contained_text_elements is None:
            contained_text_elements = set()
        self.contained_text_elements = contained_text_elements
        self.num_words = num_words
        self.num_words_in_anchor_text = num_words_in_anchor_text
        self.num_words_in_wrapped_lines = num_words_in_wrapped_lines
        self.num_wrapped_lines = num_wrapped_lines
        self.offset_blocks_start = offset_blocks
        self.offset_blocks_end = offset_blocks
        self.text_density = 0
        self.link_density = 0
        self._init_densities()
    
    def _init_densities(self) -> None:
        if self.num_words_in_wrapped_lines == 0:
            self.num_words_in_wrapped_lines = self.num_words
            self.num_wrapped_lines = 1
        self.text_density = self.num_words_in_wrapped_lines / float(self.num_wrapped_lines)
        self.link_density = 0 if self.num_words == 0 else self.num_words_in_anchor_text / float(self.num_words)
    
    def merge_next(self, next_text_block: 'TextBlock') -> None:
        if self.text is None:
            self.text = ""
        self.text = f'{self.text}\n{next_text_block.text}'
        self.num_words += next_text_block.num_words
        self.num_words_in_anchor_text += next_text_block.num_words_in_anchor_text
        self.num_words_in_wrapped_lines += next_text_block.num_words_in_wrapped_lines
        self.num_wrapped_lines += next_text_block.num_wrapped_lines
        self.offset_blocks_start = min(self.offset_blocks_start, next_text_block.offset_blocks_start)
        self.offset_blocks_end = max(self.offset_blocks_end, next_text_block.offset_blocks_end)
        self._init_densities()
        self.is_content |= next_text_block.is_content
        self.contained_text_elements |= next_text_block.contained_text_elements
        self.num_full_text_words += next_text_block.num_full_text_words
        self.labels |= next_text_block.labels
        self.tag_level = min(self.tag_level, next_text_block.tag_level)
    
    def __repr__(self) -> str:
        return f"[{self.offset_blocks_start}-{self.offset_blocks_end}; tl={self.tag_level}; nw={self.num_words};" \
               f"nwl={self.num_wrapped_lines}; ld={self.link_density}]\t" \
               f"{'CONTENT' if self.is_content else 'boilerplate'},{self.labels}\n{self.text}"
    
    def add_label(self, label) -> None:
        """
        Adds an arbitrary String label to this TextBlock.

        :param label: The label
        """
        
        self.labels.add(label)
    
    def has_label(self, label: str) -> bool:
        """
        Checks whether this TextBlock has the given label.

        :param label: The label
        :return: <code>true</code> if this block is marked by the given label.
        """
        
        return label in self.labels
    
    def remove_label(self, label: str) -> bool:
        try:
            self.labels.remove(label)
            return True
        except KeyError:
            return False
    
    def add_labels(self, *labels: Union[List, Set, Tuple]):
        """
        Adds a set of labels to this TextBlock. <code>null</code>-references are silently ignored.

        :param labels: The labels to be added.
        """
        
        if len(labels) == 0 or labels[0] is None:
            return
        if self.labels is None:
            self.labels = set()
        elif len(labels) == 1 and isinstance(labels[0], (set, list)):
            self.labels |= set(labels[0])
        else:
            self.labels |= set(labels)
    
    def clone(self):
        try:
            clone = copy.copy(self)
        except copy.error:
            raise copy.error
        if self.labels is not None:
            clone.labels = self.labels.copy()
        if self.contained_text_elements is not None:
            clone.contained_text_elements = self.contained_text_elements.copy()
        
        return clone


TextBlock.EMPTY_START = TextBlock("", set(), 0, 0, 0, 0, -1)
TextBlock.EMPTY_END = TextBlock("", set(), 0, 0, 0, 0, sys.maxsize)


class TextDocument:
    """
    A text document, consisting of one or more TextBlocks.
    """
    
    def __init__(self, text_blocks: List[TextBlock], title: str = None):
        """
        Creates a new TextDocument with given TextBlocks and given title.

        :param text_blocks: The text blocks of this document.
        :param title: The "main" title for this text document.
        """
        
        self.title = title
        self.text_blocks = text_blocks
    
    @property
    def content(self) -> str:
        """
        Returns the TextDocument's content.
        
        :return: The content text.
        """
        
        return self.get_text(True, False)
    
    def get_text(self, include_content: bool, include_non_content: bool) -> str:
        """
        Returns the TextDocument's content, non-content or both.
        
        :param include_content: Whether to include TextBlocks marked as "content".
        :param include_non_content: Whether to include TextBlocks marked as "non-content".
        :return: The text.
        """
        
        sb = []
        for block in self.text_blocks:
            if block.is_content:
                if not include_content:
                    continue
            else:
                if not include_non_content:
                    continue
            sb.append(f'{block.text}\n')
        
        return ''.join(sb)
    
    @property
    def debug_string(self) -> str:
        """
        Returns detailed debugging information about the contained TextBlocks.

        :return: Debug information.
        """
        
        return ''.join([f'{tb}\n' for tb in self.text_blocks])


class TextDocumentStatistics:
    """
    Provides shallow statistics on a given TextDocument
    """
    
    def __init__(self, doc: TextDocument, content_only: bool) -> None:
        """
        Computes statistics on a given TextDocument.
        
        :param doc: The TextDocument.
        :param content_only: if true then o
        """
        
        self.num_words = 0
        self.num_blocks = 0
        for tb in doc.text_blocks:
            if content_only and not tb.is_content:
                continue
            self.num_words += tb.num_words
            self.num_blocks += 1
    
    def avg_num_words(self) -> float:
        """
        Returns the average number of words at block-level (= overall number of words divided by the number of blocks).
        
        :return: Average
        """
        
        return self.num_words / float(self.num_blocks)



"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

from html import escape
from xml.sax.xmlreader import AttributesImpl

from boilerpy3.document import TextDocument
from boilerpy3.parser import BoilerpipeHTMLParser


class AnotherBoilerPipeHTMLParser(BoilerpipeHTMLParser):
    def __init__(self, raise_on_failure: bool = True) -> None:
        super(AnotherBoilerPipeHTMLParser, self).__init__(raise_on_failure=raise_on_failure)
    
    def error(self, message):
        pass
    
    def handle_starttag(self, tag, attributes) -> None:
        self.start_element(tag, AttributesImpl(dict(attributes)))


class HTMLBoilerpipeMarker:
    ALLOWED_ATTRIBUTES = {'class', 'href', 'src'}
    TA_IGNORABLE_ELEMENTS = {'STYLE', 'SCRIPT', 'OPTION', 'NOSCRIPT', 'OBJECT', 'EMBED', 'APPLET', 'LINK', 'HEAD',
                             'SVG', 'SELECT', 'FORM'}
    
    def __init__(self, remove_elements=None, allowed_attributes=None, raise_on_failure: bool = True) -> None:
        self.TA_IGNORABLE_ELEMENTS = set(remove_elements) if remove_elements else self.TA_IGNORABLE_ELEMENTS
        self.ALLOWED_ATTRIBUTES = set(allowed_attributes) if allowed_attributes else self.ALLOWED_ATTRIBUTES
        self.raise_on_failure = raise_on_failure
    
    def process(self, doc: TextDocument, is_: str) -> str:
        implementation = Implementation(self, raise_on_failure=self.raise_on_failure)
        implementation.process(doc, is_)
        return implementation.html


class Implementation(AnotherBoilerPipeHTMLParser):
    html = ""
    in_ignorable_element = 0
    character_element_idx = 0
    content_bit_set = set()
    
    def __init__(self, hl: HTMLBoilerpipeMarker, raise_on_failure: bool = True) -> None:
        self.hl = hl
        super(Implementation, self).__init__(raise_on_failure=raise_on_failure)
    
    def _xml_encode(self, s: str) -> str:
        return escape(s)
    
    def process(self, doc: TextDocument, is_: str):
        for block in doc.text_blocks:
            if block.is_content:
                bs = block.contained_text_elements
                if bs:
                    self.content_bit_set = self.content_bit_set.union(bs)
        
        self.feed(is_)
    
    def end_document(self):
        pass
    
    def start_document(self):
        pass
    
    def start_element(self, q_name: str, atts: dict) -> None:
        if q_name.upper() in self.hl.TA_IGNORABLE_ELEMENTS:
            self.in_ignorable_element += 1
        
        if self.in_ignorable_element == 0:
            self.html += f'<{q_name}'
            
            if self.character_element_idx + 1 in self.content_bit_set:
                self.html += ' x-boilerpipe-marker'
            
            for attr_name, attr_value in atts.items():
                if attr_name not in self.hl.ALLOWED_ATTRIBUTES:
                    continue
                self.html += f' {attr_name}=\"{self._xml_encode(attr_value or "")}\"'
            
            self.html += '>'
    
    def end_element(self, q_name: str) -> None:
        try:
            if self.in_ignorable_element == 0:
                self.html += f'</{q_name}>'
        finally:
            if q_name.upper() in self.hl.TA_IGNORABLE_ELEMENTS:
                self.in_ignorable_element -= 1
    
    def characters(self, ch: str) -> None:
        self.character_element_idx += 1
        if self.in_ignorable_element == 0:
            if self.character_element_idx not in self.content_bit_set:
                return
            
            self.html += self._xml_encode(str(ch))



"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

# -----------------------------------------------------------------------
#                           FILTER MANIFEST
# -----------------------------------------------------------------------
#
# --------------------- Simple Filters: -----------------------
# MarkEverythingContentFilter - Marks all blocks as content.
# InvertedFilter - Reverts the "is_content" flag for all TextBlocks
# BoilerplateBlockFilter - Removes TextBlocks which have explicitly been marked as "not content".
# MinWordsFilter - Keeps only those content blocks which contain at least k words.
# MinClauseWordsFilter - Keeps only blocks that have at least one segment fragment ("clause") with at least k words
# SplitParagraphBlocksFilter - Splits TextBlocks at paragraph boundaries
# SurroundingToContentFilter
# LabelToBoilerplateFilter - Marks all blocks that contain a given label as "boilerplate".
# LabelToContentFilter - Marks all blocks that contain a given label as "content".
#
# --------------------- Heuristic Filters: -----------------------
# SimpleBlockFusionProcessor - Merges two subsequent blocks if their text densities are equal.
# ContentFusion
# LabelFusion - Fuses adjacent blocks if their labels are equal.
# BlockProximityFusion - Fuses adjacent blocks if their distance (in blocks) does not exceed a certain limit.
# KeepLargestBlockFilter - Keeps the largest TextBlock only (by the number of words)
# ExpandTitleToContentFilter - Marks all TextBlocks "content" which are between the headline and the part that has
#                              already been marked content, if they are marked MIGHT_BE_CONTENT
# ArticleMetadataFilter
# AddPrecedingLabelsFilter - Adds the labels of the preceding block to the current block, optionally adding a prefix.
# DocumentTitleMatchClassifier - Marks TextBlocks which contain parts of the HTML TITLE tag
#
# --------------------- English-trained Heuristic Filters: -----------------------
# MinFulltextWordsFilter - Keeps only those content blocks which contain at least k full-text words
# KeepLargestFulltextBlockFilter - Keeps the largest TextBlock only (by the number of words)
# IgnoreBlocksAfterContentFilter - Marks all blocks as "non-content" that occur after blocks that have been marked
#                                  INDICATES_END_OF_TEXT
# IgnoreBlocksAfterContentFromEndFilter - like above
# TerminatingBlocksFinder - Finds blocks which are potentially indicating the end of an article text and marks them with
#                           INDICATES_END_OF_TEXT
# NumWordsRulesClassifier - Classifies TextBlocks as content/not-content through rules that have been determined using
#                           the C4.8 machine learning algorithm
# DensityRulesClassifier - Classifies TextBlocks as content/not-content through rules that have been determined using
#                          the C4.8 machine learning algorithm
# CanolaFilter - A full-text extractor trained on krdwrd Canola


import re
from typing import List, Pattern, Union

from boilerpy3.document import DefaultLabels, TextBlock, TextDocument


class BoilerpipeFilter:
    """
    Boilerpipe abstract interface
    """
    
    def process(self, doc: TextDocument) -> bool:
        pass
    
    def subtract_blocks(self, block_arr: List[TextBlock], blocks_to_remove: List[TextBlock]) -> List[TextBlock]:
        """
        inefficient but in place: for block in blocksToRemove: blockArr.remove(blocksToRemove) efficiently subtracts
        second array from first assuming blocksToRemove shows up in the same order as blocArr
        """
        
        if len(blocks_to_remove) == 0:
            return block_arr
        new_block_arr = []
        remove_iter = iter(blocks_to_remove)
        cur_block_to_remove = next(remove_iter)
        for idx, block in enumerate(block_arr):
            if block == cur_block_to_remove:
                try:
                    cur_block_to_remove = next(remove_iter)
                except StopIteration:
                    # add the rest
                    new_block_arr.extend(block_arr[idx + 1:])
                    break
            else:
                new_block_arr.append(block)
        return new_block_arr


class FilterChain(BoilerpipeFilter):
    """
    Chain together multiple filters in sequence
    """
    
    def __init__(self, filter_arr: List[BoilerpipeFilter]) -> None:
        super(FilterChain, self).__init__()
        self.filter_arr = filter_arr
    
    def process(self, doc: TextDocument) -> bool:
        is_updated = False
        for filtr in self.filter_arr:
            is_updated |= filtr.process(doc)
        return is_updated


# -----------------------------------------------------------------------
#                           SIMPLE FILTERS
# -----------------------------------------------------------------------


class MarkEverythingContentFilter(BoilerpipeFilter):
    """
    Marks all blocks as content.
    """
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if not tb.is_content:
                tb.is_content = True
                changes = True
        return changes


class InvertedFilter(BoilerpipeFilter):
    """
    Reverts the "is_content" flag for all TextBlocks
    """
    
    def process(self, doc: TextDocument) -> bool:
        tbs = doc.text_blocks
        if len(tbs) == 0:
            return False
        for tb in tbs:
            tb.is_content = not tb.is_content
        return True


class BoilerplateBlockFilter(BoilerpipeFilter):
    """
    Removes TextBlocks which have explicitly been marked as "not content".
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        new_blocks = [tb for tb in text_blocks if tb.is_content]
        has_changes = len(new_blocks) < len(text_blocks)
        doc.text_blocks = new_blocks
        
        return has_changes


class MinWordsFilter(BoilerpipeFilter):
    """
    Keeps only those content blocks which contain at least <em>k</em> words.
    """
    
    def __init__(self, min_words: int) -> None:
        super(MinWordsFilter, self).__init__()
        self.min_words = min_words
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if not tb.is_content:
                continue
            if tb.num_words < self.min_words:
                tb.is_content = False
                changes = True
        return changes


class MinClauseWordsFilter(BoilerpipeFilter):
    """
    Keeps only blocks that have at least one segment fragment ("clause") with at least <em>k</em> words (default: 5).

    NOTE: You might consider using the SplitParagraphBlocksFilter
    upstream.

    See SplitParagraphBlocksFilter
    """
    
    PAT_CLAUSE_DELIMITER = re.compile(r"\b[,.:;!?]+(?:\s+|\Z)", re.UNICODE)
    PAT_WHITESPACE = re.compile(r"\s+")
    
    def __init__(self, min_words: int = 5, accept_clauses_without_delimiter: bool = False) -> None:
        super(MinClauseWordsFilter, self).__init__()
        self.min_words = min_words
        self.accept_clauses_without_delimiter = accept_clauses_without_delimiter
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if not tb.is_content:
                continue
            has_clause = False
            possible_clause_arr = self.PAT_CLAUSE_DELIMITER.split(tb.text)
            for possible_clause in possible_clause_arr[:-1]:
                has_clause = self.is_clause_accepted(possible_clause)
                if has_clause:
                    break
            
            # since clauses should *always end* with a delimiter, we normally don't consider text without one
            if self.accept_clauses_without_delimiter:
                has_clause |= self.is_clause_accepted(possible_clause_arr[-1])
            if not has_clause:
                tb.is_content = False
                changes = True
        return changes
    
    def is_clause_accepted(self, text: str):
        n = 1
        for _ in self.PAT_WHITESPACE.finditer(text):
            n += 1
            if n >= self.min_words:
                return True
        return n >= self.min_words


class SplitParagraphBlocksFilter(BoilerpipeFilter):
    """
    Splits TextBlocks at paragraph boundaries.

    NOTE: This is not fully supported (i.e., it will break highlighting support via #getContainedTextElements()), but
    this one probably is necessary for some other filters.

    See MinClauseWordsFilter
    """
    
    NEWLINE_REGEX = re.compile(r"[\n\r]+")
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        blocks = doc.text_blocks
        blocks_new = []
        for tb in blocks:
            text = tb.text
            paragraphs = self.NEWLINE_REGEX.split(text)
            if len(paragraphs) < 2:
                blocks_new.append(tb)
                continue
            is_content = tb.is_content
            labels = tb.labels
            for p in paragraphs:
                tb_p = TextBlock(p)
                tb_p.is_content = is_content
                tb_p.add_labels(labels)
                blocks_new.append(tb_p)
                changes = True
        
        if changes:
            doc.text_blocks = blocks_new
        return changes


class SurroundingToContentFilter(BoilerpipeFilter):
    def __init__(self, condition: callable = lambda tb: tb.linkDensity == 0 and tb.num_words > 6) -> None:
        """
        this is now default when no arguments are passed
        
        INSTANCE_TEXT = SurroundingToContentFilter(TextBlockCondition())
        
        ctor - condition is an function for an additional condition to determine if it can be made content
        """
        
        super(SurroundingToContentFilter, self).__init__()
        self.cond = condition
    
    def process(self, doc: TextDocument) -> bool:
        tbs = doc.text_blocks
        n = len(tbs)
        has_changes = False
        i = 1
        while i < n - 1:
            prev_block = tbs[i - 1]
            cur_block = tbs[i]
            next_block = tbs[i + 1]
            if not cur_block.is_content and prev_block.is_content and next_block.is_content and self.cond(cur_block):
                cur_block.is_content = True
                has_changes = True
                i += 2
            else:
                # WARNING: POSSIBLE BUG - in original i+=2 regardless of whether content is found. this seems illogical
                # to me - should be +=1
                i += 1
        
        return has_changes


class LabelToBoilerplateFilter(BoilerpipeFilter):
    """
    Marks all blocks that contain a given label as "boilerplate".
    
    INSTANCE_STRICTLY_NOT_CONTENT = LabelToBoilerplateFilter(DefaultLabels.STRICTLY_NOT_CONTENT)
    """
    
    def __init__(self, *labels: str) -> None:
        super(LabelToBoilerplateFilter, self).__init__()
        self.labels = labels
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if tb.is_content and any(tb.has_label(label) for label in self.labels):
                tb.is_content = False
                changes = True
        return changes


class LabelToContentFilter(BoilerpipeFilter):
    """
    Marks all blocks that contain a given label as "content".
    """
    
    def __init__(self, *labels: str) -> None:
        super(LabelToContentFilter, self).__init__()
        self.labels = labels
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if not tb.is_content and any(tb.has_label(label) for label in self.labels):
                tb.is_content = True
                changes = True
        return changes


# -----------------------------------------------------------------------
#                       GENERIC HEURISTIC FILTERS
# -----------------------------------------------------------------------


class SimpleBlockFusionProcessor(BoilerpipeFilter):
    """
    Merges two subsequent blocks if their text densities are equal.
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        changes = False
        if len(text_blocks) < 2:
            return False
        prev_block = text_blocks[0]
        blocks_to_remove = []
        for block in text_blocks[1:]:
            if prev_block.text_density == block.text_density:
                prev_block.merge_next(block)
                blocks_to_remove.append(block)
                changes = True
            else:
                prev_block = block
        
        if changes:
            doc.text_blocks = self.subtract_blocks(text_blocks, blocks_to_remove)
        
        return changes


class ContentFusion(BoilerpipeFilter):
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        # WARNING: POSSIBLE BUG FOUND: shouldn't prev_block be reset every passthrough?
        changes = False
        # if it has been changed on the previous passthrough
        changed_on_pass = True
        while changed_on_pass:
            changed_on_pass = False
            prev_block = text_blocks[0]
            blocks_to_remove = []
            for block in text_blocks[1:]:
                if prev_block.is_content and block.link_density < 0.56 \
                        and not block.has_label(DefaultLabels.STRICTLY_NOT_CONTENT):
                    prev_block.merge_next(block)
                    blocks_to_remove.append(block)
                    changed_on_pass = True
                    changes = True
                else:
                    prev_block = block
                text_blocks = self.subtract_blocks(text_blocks, blocks_to_remove)
        if changes:
            doc.text_blocks = text_blocks
        
        return changes


class LabelFusion(BoilerpipeFilter):
    """
    Fuses adjacent blocks if their labels are equal.
    """
    
    def __init__(self, label_prefix: str = "") -> None:
        """
        Creates a new LabelFusion instance.
        
        :param label_prefix: The maximum distance in blocks.
        """
        
        super(LabelFusion, self).__init__()
        self.label_prefix = label_prefix
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        changes = False
        prev_block = text_blocks[0]
        blocks_to_remove = []
        for block in text_blocks[1::]:
            if self.equal_labels(prev_block.labels, block.labels):
                prev_block.merge_next(block)
                blocks_to_remove.append(block)
                changes = True
            else:
                prev_block = block
        
        if changes:
            doc.text_blocks = self.subtract_blocks(text_blocks, blocks_to_remove)
        
        return changes
    
    def equal_labels(self, labels1: List[str], labels2: List[str]) -> bool:
        if labels1 is None or labels2 is None:
            return False
        
        # NOTE: Should blocks be merged if neither of them have labels???  i.e. labels1==labels2==empty set
        return self.markup_labels_only(labels1) == self.markup_labels_only(labels2)
    
    def markup_labels_only(self, labels: List[str]) -> set:
        return {label for label in labels if label.startswith(DefaultLabels.MARKUP_PREFIX)}


class BlockProximityFusion(BoilerpipeFilter):
    """
    Fuses adjacent blocks if their distance (in blocks) does not exceed a certain limit. This probably makes sense only
    in cases where an upstream filter already has removed some blocks.
    
    MAX_DISTANCE_1 = BlockProximityFusion(1, False, False)
    MAX_DISTANCE_1_SAME_TAGLEVEL = BlockProximityFusion(1, False, True)
    MAX_DISTANCE_1_CONTENT_ONLY = BlockProximityFusion(1, True, False)
    MAX_DISTANCE_1_CONTENT_ONLY_SAME_TAGLEVEL = BlockProximityFusion(1, True, True)
    """
    
    def __init__(self, max_blocks_distance: int = 1, content_only: bool = False,
                 same_tag_level_only: bool = False) -> None:
        """
        Creates a new BlockProximityFusion instance.
        
        :param max_blocks_distance: The maximum distance in blocks.
        :param content_only:
        :param same_tag_level_only:
        """
        
        super(BlockProximityFusion, self).__init__()
        self.max_blocks_distance = max_blocks_distance
        self.content_only = content_only
        self.same_tag_level_only = same_tag_level_only
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        changes = False
        
        if self.content_only:
            start_idx = None
            for idx, block in enumerate(text_blocks):
                if block.is_content:
                    start_idx = idx
                    break
            if start_idx is None:
                return False
        else:
            start_idx = 0
        
        prev_block = text_blocks[start_idx]
        blocks_to_remove = []
        for block in text_blocks[start_idx + 1:]:
            if not block.is_content:
                prev_block = block
                continue
            diff_blocks = block.offset_blocks_start - prev_block.offset_blocks_end - 1
            if diff_blocks <= self.max_blocks_distance:
                ok = True
                if self.content_only:
                    if not prev_block.is_content or not block.is_content:
                        ok = False
                if self.same_tag_level_only and prev_block.tag_level != block.tag_level:
                    ok = False
                if ok:
                    prev_block.merge_next(block)
                    # remove current block
                    blocks_to_remove.append(block)
                    changes = True
                else:
                    prev_block = block
            else:
                prev_block = block
        
        if len(blocks_to_remove) > 0:
            doc.text_blocks = self.subtract_blocks(text_blocks, blocks_to_remove)
            changes = True
        
        return changes


class KeepLargestBlockFilter(BoilerpipeFilter):
    """
    Keeps the largest TextBlock only (by the number of words). In case of more than one block with the same
    number of words, the first block is chosen. All discarded blocks are marked "not content" and flagged as
    DefaultLabels.

    Note that, by default, only TextBlocks marked as "content" are taken into consideration.
    
    INSTANCE = KeepLargestBlockFilter(False)
    INSTANCE_EXPAND_TO_SAME_TAGLEVEL = KeepLargestBlockFilter(True)
    """
    
    def __init__(self, expand_to_same_level_text: bool = False) -> None:
        super(KeepLargestBlockFilter, self).__init__()
        self.expand_to_same_level_text = expand_to_same_level_text
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        
        try:
            largest_block = max((tb for tb in text_blocks if tb.is_content), key=lambda tb: tb.num_words)
        except ValueError:
            # no content blocks exist / largest block not found
            largest_block = None
        
        for tb in text_blocks:
            if tb == largest_block:
                tb.is_content = True
            else:
                tb.is_content = False
                tb.add_label(DefaultLabels.MIGHT_BE_CONTENT)
        
        if self.expand_to_same_level_text and largest_block is not None:
            level = largest_block.tag_level
            largest_block_idx = text_blocks.index(largest_block)
            
            for tb in text_blocks[largest_block_idx::-1]:
                tl = tb.tag_level
                if tl < level:
                    break
                elif tl == level:
                    tb.is_content = True
            
            for tb in text_blocks[largest_block_idx:]:
                tl = tb.tag_level
                if tl < level:
                    break
                elif tl == level:
                    tb.is_content = True
        
        return True


class ExpandTitleToContentFilter(BoilerpipeFilter):
    """
    Marks all TextBlocks "content" which are between the headline and the part that has already been marked
    content, if they are marked DefaultLabels#MIGHT_BE_CONTENT.
    
    This filter is quite specific to the news domain.
    """
    
    def process(self, doc: TextDocument) -> bool:
        i = 0
        title_idx = -1
        content_start = -1
        for tb in doc.text_blocks:
            if content_start == -1 and tb.has_label(DefaultLabels.TITLE):
                title_idx = i
            if content_start == -1 and tb.is_content:
                content_start = i
            i += 1
        
        if content_start <= title_idx or title_idx == -1:
            return False
        
        changes = False
        for tb in doc.text_blocks[title_idx:content_start]:
            if tb.has_label(DefaultLabels.MIGHT_BE_CONTENT):
                if tb.is_content is not True:
                    tb.is_content = True
                    changes = True
        
        return changes


class ArticleMetadataFilter(BoilerpipeFilter):
    # checks for date/time/author blocks
    PATTERNS_SHORT = [
        re.compile(r"^[0-9 ,./]*\b(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec|January|February|March|April|June|"
                   r"July|August|September|October|November|December)?\b[0-9 ,:apm./]*(?:[CPSDMGET]{2,3})?$"),
        re.compile("^[Bb]y ")
    ]
    
    def process(self, doc: TextDocument) -> bool:
        changed = False
        for tb in doc.text_blocks:
            if tb.num_words > 10:
                continue
            for p in self.PATTERNS_SHORT:
                text = tb.text
                if p.search(text):
                    changed = True
                    tb.is_content = True
                    tb.add_label(DefaultLabels.ARTICLE_METADATA)
                    break
        return changed


class AddPrecedingLabelsFilter(BoilerpipeFilter):
    """
    Adds the labels of the preceding block to the current block, optionally adding a prefix.
    """
    
    def __init__(self, label_prefix: str = "") -> None:
        """
        Creates a new AddPrecedingLabelsFilter instance.
        
        INSTANCE = AddPrecedingLabelsFilter("")
        INSTANCE_PRE = AddPrecedingLabelsFilter("^")
        """
        
        super(AddPrecedingLabelsFilter, self).__init__()
        self.label_prefix = label_prefix
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        changes = False
        block_below = None
        
        for block in text_blocks[::-1]:
            if block_below is not None:
                labels = block.labels
                if labels is not None and len(labels) > 0:
                    for l in labels:
                        block_below.add_label(self.label_prefix + l)
                    changes = True
            block_below = block
        
        return changes


class DocumentTitleMatchClassifier(BoilerpipeFilter):
    """
    Marks TextBlocks which contain parts of the HTML <code>&lt;TITLE&gt;</code> tag, using some heuristics which
    are quite specific to the news domain.
    """
    
    TITLE_REGEXES = [
        re.compile(r"[ ]*[|:][ ]*"),
        re.compile(r"[ ]*[|:()][ ]*"),
        re.compile(r"[ ]*[|:()\-][ ]*"),
        re.compile(r"[ ]*[|,:()\-][ ]*")
    ]
    WORD_REGEX = re.compile(r"\w+", re.UNICODE)
    
    def __init__(self, title: Union[str, None], use_doc_title: bool = False) -> None:
        super(DocumentTitleMatchClassifier, self).__init__()
        self.use_doc_title = use_doc_title
        if use_doc_title:
            self.potential_titles = None
        else:
            self.potential_titles = self.find_potential_titles(title)
    
    def find_potential_titles(self, title: str):
        if title is None:
            return None
        title = title.strip()
        if len(title) == 0:
            return None
        else:
            potential_titles = set()
            potential_titles.add(title)
            for regex in self.TITLE_REGEXES:
                p = self.get_longest_part(title, regex)
                if p is not None:
                    potential_titles.add(p)
        return potential_titles
    
    def get_longest_part(self, title: str, pattern: Pattern):
        parts = pattern.split(title)
        if len(parts) == 1:
            return None
        
        longest_num_words = 0
        longest_part = ""
        for p in parts:
            if ".com" in p:
                continue
            num_words = self.get_num_words(p)
            if num_words > longest_num_words or len(p) > len(longest_part):
                longest_num_words = num_words
                longest_part = p
        if len(longest_part) == 0:
            return None
        else:
            return longest_part.strip()
    
    def get_num_words(self, text: str):
        return len(self.WORD_REGEX.findall(text))
    
    def process(self, doc: TextDocument) -> bool:
        if self.use_doc_title:
            self.potential_titles = self.find_potential_titles(doc.title)
        if self.potential_titles is None:
            return False
        changes = False
        for tb in doc.text_blocks:
            text = tb.text.strip().lower()
            if any(candidate.lower() == text for candidate in self.potential_titles):
                tb.add_label(DefaultLabels.TITLE)
                changes = True
        return changes


# -----------------------------------------------------------------------
#                          ENGLISH HEURISTIC FILTERS
# -----------------------------------------------------------------------
# --- Heuristic Filters that have been trained on English laguage text


class HeuristicFilterBase(BoilerpipeFilter):
    """
    Base class for some heuristics that are used by boilerpipe filters.
    """
    
    def get_num_full_text_words(self, tb: TextBlock, min_text_density: int = 9):
        if tb.text_density >= min_text_density:
            return tb.num_words
        else:
            return 0


class MinFulltextWordsFilter(HeuristicFilterBase):
    """
    Keeps only those content blocks which contain at least k full-text words (measured by
    HeuristicFilterBase#get_num_full_text_words(TextBlock). k is 30 by default.
    """
    
    def __init__(self, min_words: int = 30) -> None:
        self.min_words = min_words
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        for tb in doc.text_blocks:
            if tb.is_content and self.get_num_full_text_words(tb) < self.min_words:
                tb.is_content = False
                changes = True
        return changes


class KeepLargestFulltextBlockFilter(HeuristicFilterBase):
    """
    Keeps the largest TextBlock only (by the number of words). In case of more than one block with the same
    number of words, the first block is chosen. All discarded blocks are marked "not content" and flagged as
    DefaultLabels. As opposed to KeepLargestBlockFilter, the number of words are computed using HeuristicFilterBase
    get_num_full_text_words(TextBlock), which only counts words that occur in text elements with at least 9 words and
    are thus believed to be full text.

    NOTE: Without language-specific fine-tuning (i.e., running the default instance), this filter may lead to suboptimal
    results. You better use KeepLargestBlockFilter instead, which works at the level of number-of-words instead
    of text densities.
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        if len(text_blocks) < 2:
            return False
        content_blocks = [block for block in text_blocks if block.is_content]
        if len(content_blocks) == 0:
            return False
        largest_block = max(content_blocks, key=self.get_num_full_text_words)
        
        for tb in text_blocks:
            if tb == largest_block:
                tb.is_content = True
            else:
                tb.is_content = False
                tb.add_label(DefaultLabels.MIGHT_BE_CONTENT)
        return True


class IgnoreBlocksAfterContentFilter(HeuristicFilterBase):
    """
    Marks all blocks as "non-content" that occur after blocks that have been marked DefaultLabels#INDICATES_END_OF_TEXT.
    These marks are ignored unless a minimum number of words in content blocks occur before this mark (default: 60).
    This can be used in conjunction with an upstream TerminatingBlocksFinder.
    """
    
    def __init__(self, min_num_words: int = 60) -> None:
        """
        DEFAULT_INSTANCE = IgnoreBlocksAfterContentFilter(60)
        INSTANCE_200 = IgnoreBlocksAfterContentFilter(200)
        """
        
        self.min_num_words = min_num_words
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        num_words = 0
        found_end_of_text = False
        for block in doc.text_blocks:
            if block.is_content:
                num_words += self.get_num_full_text_words(block)
            if block.has_label(DefaultLabels.INDICATES_END_OF_TEXT) and num_words >= self.min_num_words:
                found_end_of_text = True
            if found_end_of_text:
                changes = True
                block.is_content = False
        
        return changes


class IgnoreBlocksAfterContentFromEndFilter(HeuristicFilterBase):
    """
    Marks all blocks as "non-content" that occur after blocks that have been marked DefaultLabels#INDICATES_END_OF_TEXT,
    and after any content block. This filter can be used in conjunction with an upstream TerminatingBlocksFinder.
    
    See TerminatingBlocksFinder
    """
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        words = 0
        blocks = doc.text_blocks
        if len(blocks) == 0:
            return False
        for tb in blocks[::-1]:
            if tb.has_label(DefaultLabels.INDICATES_END_OF_TEXT):
                tb.add_label(DefaultLabels.STRICTLY_NOT_CONTENT)
                tb.remove_label(DefaultLabels.MIGHT_BE_CONTENT)
                tb.is_content = False
                changes = True
            elif tb.is_content:
                words += tb.num_words
                if words > 200:
                    break
        return changes


class TerminatingBlocksFinder(BoilerpipeFilter):
    """
    Finds blocks which are potentially indicating the end of an article text and marks them with
    DefaultLabels#INDICATES_END_OF_TEXT. This can be used in conjunction with a downstream
    IgnoreBlocksAfterContentFilter.
    """
    
    DIGIT_REGEX = re.compile(r'\D')
    
    def process(self, doc: TextDocument) -> bool:
        changes = False
        
        for tb in doc.text_blocks:
            if tb.num_words >= 15:
                continue
            text = tb.text.strip()
            if len(text) < 8:
                continue
            text_lc = text.lower()
            
            startmatches = (" reuters", "please rate this", "post a comment")
            inmatches = ("what you think...", "add your comment", "add comment", "reader views", "have your say",
                         "reader comments", "rtta artikeln")
            eqmatch = "thanks for your comments - this feedback is now closed"
            
            if text_lc.startswith("comments") or self.starts_with_number(text_lc, " comments", " users responded in") \
                    or any(text_lc.startswith(match_str) for match_str in startmatches) \
                    or any(match_str in text_lc for match_str in inmatches) or text_lc == eqmatch:
                tb.add_label(DefaultLabels.INDICATES_END_OF_TEXT)
                changes = True
        
        return changes
    
    def starts_with_number(self, text: str, *match_str_arr: str):
        """
        Checks whether the given text t starts with a sequence of digits, followed by one of the given strings.
        
        :param text: The text to examine
        :param match_str_arr: Any strings that may follow the digits.
        :return: true if at least one combination matches
        """
        
        number_match = self.DIGIT_REGEX.search(text)
        if number_match is None:
            pos = len(text)
        else:
            pos = number_match.start()
        if pos == 0:
            return False
        else:
            return any(text.startswith(match_str, pos) for match_str in match_str_arr)


class NumWordsRulesClassifier(BoilerpipeFilter):
    """
    Classifies TextBlocks as content/not-content through rules that have been determined using the C4.8 machine
    learning algorithm, as described in the paper "Boilerplate Detection using Shallow Text Features" (WSDM 2010),
    particularly using number of words per block and link density per block.
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        has_changes = False
        
        n = len(text_blocks)
        for i, currentBlock in enumerate(text_blocks):
            if i > 0:
                prev_block = text_blocks[i - 1]
            else:
                prev_block = TextBlock.EMPTY_START
            if i + 1 < n:
                next_block = text_blocks[i + 1]
            else:
                next_block = TextBlock.EMPTY_START
            has_changes |= self.classify(prev_block, currentBlock, next_block)
        return has_changes
    
    def classify(self, prev_block: TextBlock, curr_block: TextBlock, next_block: TextBlock):
        if curr_block.link_density <= 0.333333:
            if prev_block.link_density <= 0.555556:
                if curr_block.num_words <= 16:
                    if next_block.num_words <= 15:
                        if prev_block.num_words <= 4:
                            is_content = False
                        else:
                            is_content = True
                    else:
                        is_content = True
                else:
                    is_content = True
            else:
                if curr_block.num_words <= 40:
                    if next_block.num_words <= 17:
                        is_content = False
                    else:
                        is_content = True
                else:
                    is_content = True
        else:
            is_content = False
        
        changes = curr_block.is_content is is_content
        curr_block.is_content = is_content
        
        return changes


class DensityRulesClassifier(BoilerpipeFilter):
    """
    Classifies TextBlocks as content/not-content through rules that have been determined using the C4.8 machine learning
    algorithm, as described in the paper "Boilerplate Detection using Shallow Text Features", particularly using text
    densities and link densities.
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        has_changes = False
        
        n = len(text_blocks)
        for i, current_block in enumerate(text_blocks):
            if i > 0:
                prev_block = text_blocks[i - 1]
            else:
                prev_block = TextBlock.EMPTY_START
            if i + 1 < n:
                next_block = text_blocks[i + 1]
            else:
                next_block = TextBlock.EMPTY_START
            has_changes |= self.classify(prev_block, current_block, next_block)
        return has_changes
    
    def classify(self, prev_block: TextBlock, curr_block: TextBlock, next_block: TextBlock):
        if curr_block.link_density <= 0.333333:
            if prev_block.link_density <= 0.555556:
                if curr_block.text_density <= 9:
                    if next_block.text_density <= 10:
                        if prev_block.text_density <= 4:
                            is_content = False
                        else:
                            is_content = True
                    else:
                        is_content = True
                else:
                    if next_block.text_density == 0:
                        is_content = False
                    else:
                        is_content = True
            else:
                if next_block.text_density <= 11:
                    is_content = False
                else:
                    is_content = True
        else:
            is_content = False
        
        changes = curr_block.is_content is is_content
        curr_block.is_content = is_content
        
        return changes


class CanolaFilter(BoilerpipeFilter):
    """
    A full-text extractor trained on http://krdwrd.org/,
    https://krdwrd.org/trac/attachment/wiki/Corpora/Canola/CANOLA.pdf. Works well with SimpleEstimator, too.
    """
    
    def process(self, doc: TextDocument) -> bool:
        text_blocks = doc.text_blocks
        has_changes = False
        
        n = len(text_blocks)
        for i, current_block in enumerate(text_blocks):
            if i > 0:
                prev_block = text_blocks[i - 1]
            else:
                prev_block = TextBlock.EMPTY_START
            if i + 1 < n:
                next_block = text_blocks[i + 1]
            else:
                next_block = TextBlock.EMPTY_START
            has_changes |= self.classify(prev_block, current_block, next_block)
        return has_changes
    
    def classify(self, prev_block: TextBlock, curr_block: TextBlock, next_block: TextBlock):
        cond1 = curr_block.link_density > 0 and next_block.num_words > 11
        cond2 = curr_block.num_words > 19
        cond3 = next_block.num_words > 6 and next_block.link_density == 0 and prev_block.link_density == 0 and \
                (curr_block.num_words > 6 or prev_block.num_words > 7 or next_block.num_words > 19)
        is_content = cond1 or cond2 or cond3
        
        changes = curr_block.is_content is is_content
        curr_block.is_content = is_content
        
        return changes



"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

import re
from html.parser import HTMLParser
from logging import getLogger
from typing import Dict, Set
from xml.sax.handler import ContentHandler
from xml.sax.xmlreader import AttributesImpl

from boilerpy3.document import DefaultLabels, TextBlock, TextDocument

logger = getLogger('boilerpy3')


# ----------------------------------------------------------------------------
#                                TAG ACTIONS
# ----------------------------------------------------------------------------


class TagAction:
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        return False
    
    def changes_tag_level(self) -> bool:
        return False


class IgnorableElementTagAction(TagAction):
    """
    Marks this tag as "ignorable", i.e. all its inner content is silently skipped.
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.in_ignorable_element += 1
        return True
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.in_ignorable_element -= 1
        return True
    
    def changes_tag_level(self) -> bool:
        return True


class AnchorTextTagAction(TagAction):
    """
    Marks this tag as "anchor" (this should usually only be set for the <code>&lt;A&gt;</code> tag). Anchor tags may not
    be nested.
    
    There is a bug in certain versions of NekoHTML which still allows nested tags. If boilerpipe encounters such
    nestings, a SAXException is thrown.
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.in_anchor += 1
        if content_handler.in_anchor > 1:
            # as nested A elements are not allowed per specification, we are probably reaching this branch due to a bug
            # in the XML parser
            logger.warning("Warning: SAX input contains nested A elements -- You have probably hit a bug in your HTML "
                           "parser (e.g., NekoHTML bug #2909310). Please clean the HTML externally and feed it to "
                           "BoilerPy3 again. Trying to recover somehow...")
            self.end(content_handler, tag_name)
        if content_handler.in_ignorable_element == 0:
            content_handler.add_token(SpecialTokens.ANCHOR_TEXT_START)
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.in_anchor -= 1
        if content_handler.in_anchor == 0 and content_handler.in_ignorable_element == 0:
            content_handler.add_token(SpecialTokens.ANCHOR_TEXT_END)
        return False
    
    def changes_tag_level(self) -> bool:
        return True


class BodyTagAction(TagAction):
    """
    Marks this tag the body element (this should usually only be set for the <code>&lt;BODY&gt;</code> tag).
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.flush_block()
        content_handler.in_body += 1
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.flush_block()
        content_handler.in_body -= 1
        return False
    
    def changes_tag_level(self) -> bool:
        return True


class InlineWhitespaceTagAction(TagAction):
    """
    Marks this tag a simple "inline" element, which generates whitespace, but no new block.
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.add_whitespace_if_necessary()
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.add_whitespace_if_necessary()
        return False
    
    def changes_tag_level(self) -> bool:
        return False


class InlineTagAction(TagAction):
    """
    Marks this tag a simple "inline" element, which neither generates whitespace, nor a new block.
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl):
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        return False
    
    def changes_tag_level(self) -> bool:
        return False


class BlockTagAction(TagAction):
    """
    Explicitly marks this tag a simple "block-level" element, which always generates whitespace
    """
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        return True
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        return True
    
    def changes_tag_level(self) -> bool:
        return True


class FontTagAction(TagAction):
    """
    Special TagAction for the <code>&lt;FONT&gt;</code> tag, which keeps track of the absolute and relative font size.
    """
    
    # WARNING: POSSIBLE BUG -- used to be [0-9] without +
    PAT_FONT_SIZE = re.compile(r"([+\-]?)([0-9]+)")
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        size_attr = attrs.getValue("size")
        size = None
        if size_attr is not None:
            match = self.PAT_FONT_SIZE.match(size_attr)
            if match is not None:
                rel = match.group(0)
                val = match.group(1)
                # absolute
                if len(rel) == 0:
                    size = val
                # relative
                else:
                    # last non-none element from stack, default 3
                    last_non_none = (s for s in content_handler.font_size_stack[::-1] if s is not None)
                    prev_size = next(last_non_none, 3)
                    if rel[0] == '+':
                        size = prev_size + val
                    else:
                        size = prev_size - val
        content_handler.font_size_stack.append(size)
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.font_size_stack.pop()
        return False
    
    def changes_tag_level(self) -> bool:
        return False


class InlineTagLabelAction(TagAction):
    """
    CommonTagActions for inline elements, which triggers some LabelAction on the generated TextBlock.
    """
    
    def __init__(self, action) -> None:
        super(InlineTagLabelAction, self).__init__()
        self.action = action
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.add_whitespace_if_necessary()
        content_handler.add_label_action(self.action)
        return False
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        content_handler.add_whitespace_if_necessary()
        return False
    
    def changes_tag_level(self) -> bool:
        return False


class BlockTagLabelAction(TagAction):
    """
    CommonTagActions for block-level elements, which triggers some LabelAction} on the generated TextBlock.
    """
    
    def __init__(self, action: 'LabelAction') -> None:
        super(BlockTagLabelAction, self).__init__()
        self.action = action
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        content_handler.add_label_action(self.action)
        return True
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        return True
    
    def changes_tag_level(self) -> bool:
        return True


class Chained(TagAction):
    def __init__(self, tag_action1: TagAction, tag_action2: TagAction) -> None:
        super(Chained, self).__init__()
        self.tag_action1 = tag_action1
        self.tag_action2 = tag_action2
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        return self.tag_action1.start(content_handler, tag_name, attrs) | \
               self.tag_action2.start(content_handler, tag_name, attrs)
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        return self.tag_action1.end(content_handler, tag_name) | self.tag_action2.end(content_handler, tag_name)
    
    def changes_tag_level(self) -> bool:
        return self.tag_action1.changes_tag_level() or self.tag_action2.changes_tag_level()


class MarkupTagAction(TagAction):
    PAT_NUM = re.compile("[0-9]+")
    
    def __init__(self, is_block_level: bool) -> None:
        super(MarkupTagAction, self).__init__()
        self.is_block_level = is_block_level
        self.label_stack = []
    
    def start(self, content_handler: 'BoilerpipeBaseParser', tag_name: str, attrs: AttributesImpl) -> bool:
        labels = [DefaultLabels.MARKUP_PREFIX + tag_name]
        class_val = attrs.getValue("class")
        if class_val is not None and len(class_val) > 0:
            class_val = self.PAT_NUM.sub("#", class_val).strip()
            vals = class_val.split(r"[ ]+")
            labels.append(f"{DefaultLabels.MARKUP_PREFIX}.{class_val.replace(' ', '.')}")
            if len(vals) > 1:
                for s in vals:
                    labels.append(f"{DefaultLabels.MARKUP_PREFIX}.{s}")
        block_id = attrs.get("id")
        if block_id is not None and len(block_id) < 0:
            block_id = self.PAT_NUM.sub("#", block_id)
            labels.append(f"{DefaultLabels.MARKUP_PREFIX}#{block_id}")
        ancestors = self.get_ancestor_labels()
        labels_with_ancestors = []
        for l in labels:
            for an in ancestors:
                labels_with_ancestors.append(an)
                labels_with_ancestors.append(f"{an} {l}")
            labels_with_ancestors.append(l)
        content_handler.add_label_action(LabelAction(*labels_with_ancestors))
        self.label_stack.append(labels)
        return self.is_block_level
    
    def end(self, content_handler: 'BoilerpipeBaseParser', tag_name: str) -> bool:
        self.label_stack.pop()
        return self.is_block_level
    
    def changes_tag_level(self) -> bool:
        return self.is_block_level
    
    def get_ancestor_labels(self) -> Set[str]:
        label_set = set()
        for labels in label_set:
            if labels is None:
                continue
            label_set.update(labels)
        return label_set


class CommonTagActions:
    TA_IGNORABLE_ELEMENT = IgnorableElementTagAction()
    TA_ANCHOR_TEXT = AnchorTextTagAction()
    TA_BODY = BodyTagAction()
    TA_INLINE_WHITESPACE = InlineWhitespaceTagAction()
    TA_INLINE_NO_WHITESPACE = InlineTagAction()
    TA_BLOCK_LEVEL = BlockTagAction()
    TA_FONT = FontTagAction()


defaultTagActionMap = {
    "STYLE": CommonTagActions.TA_IGNORABLE_ELEMENT,
    "SCRIPT": CommonTagActions.TA_IGNORABLE_ELEMENT,
    "OPTION": CommonTagActions.TA_IGNORABLE_ELEMENT,
    "OBJECT": CommonTagActions.TA_IGNORABLE_ELEMENT,
    "EMBED": CommonTagActions.TA_IGNORABLE_ELEMENT,
    "APPLET": CommonTagActions.TA_IGNORABLE_ELEMENT,
    # Note: link removed because it can be self-closing in HTML5
    # "LINK" : CommonTagActions.TA_IGNORABLE_ELEMENT,
    "A": CommonTagActions.TA_ANCHOR_TEXT,
    "BODY": CommonTagActions.TA_BODY,
    "STRIKE": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "U": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "B": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "I": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "EM": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "STRONG": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "SPAN": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    #  New in 1.1 (especially to improve extraction quality from Wikipedia etc.,
    "SUP": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    #  New in 1.2
    "CODE": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "TT": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "SUB": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "VAR": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    "ABBR": CommonTagActions.TA_INLINE_WHITESPACE,
    "ACRONYM": CommonTagActions.TA_INLINE_WHITESPACE,
    "FONT": CommonTagActions.TA_INLINE_NO_WHITESPACE,
    #  could also use TA_FONT
    #  added in 1.1.1
    "NOSCRIPT": CommonTagActions.TA_IGNORABLE_ELEMENT
}


# ----------------------------------------------------------------------------
#                                LABEL ACTIONS
# ----------------------------------------------------------------------------

class LabelAction:
    """
    Helps adding labels to TextBlocks.
    """
    
    def __init__(self, *labels: str) -> None:
        self.labels = labels
    
    def add_to(self, text_block: TextBlock) -> None:
        self.add_labels_to(text_block)
    
    def add_labels_to(self, text_block: TextBlock):
        text_block.add_labels(self.labels)
    
    def __str__(self):
        return str(self.labels)


class ConditionalLabelAction(LabelAction):
    def __init__(self, condition, *labels: str):
        super(ConditionalLabelAction, self).__init__(*labels)
        self.condition = condition
    
    def add_to(self, text_block: TextBlock):
        if self.condition(text_block):
            self.add_labels_to(text_block)


class SpecialTokens:
    ANCHOR_TEXT_START = '\ue00astart'
    ANCHOR_TEXT_END = '\ue00aend'


# ----------------------------------------------------------------------------
#                           SAX CONTENT HANDLER
# ----------------------------------------------------------------------------


class BoilerpipeBaseParser:
    """
    A simple SAX ContentHandler, used by BoilerpipeSAXInput. Can be used by different parser
    implementations, e.g. NekoHTML and TagSoup.
    """
    
    EVENT_START_TAG = 0
    EVENT_END_TAG = 1
    EVENT_CHARACTERS = 2
    EVENT_WHITESPACE = 3
    # all word characters except underscore -- i.e. not (not word or underscore)
    PAT_VALID_WORD_CHARACTER = re.compile(r"[^\W_]", re.UNICODE)
    PAT_WORD = re.compile(r"\ue00a?[\w\"'.,!@\-:;$?()/]+", re.UNICODE)
    
    def __init__(self, tag_actions: Dict[str, TagAction] = None, raise_on_failure: bool = True) -> None:
        """
        Constructs a BoilerpipeHTMLContentHandler using the given TagActionMap.
        
        :param tag_actions: The TagActionMap to use, e.g. DefaultTagActionMap.
        """
        
        if tag_actions is None:
            self.tagActions = defaultTagActionMap
        else:
            self.tagActions = tag_actions
        
        self.clear_text_buffer()
        self.in_body = 0
        self.in_anchor = 0
        self.in_ignorable_element = 0
        self.text_element_idx = 0
        self.last_start_tag = None
        self.last_end_tag = None
        self.last_event = None
        self.offset_blocks = 0
        self.current_contained_text_elements = set()
        self.flush = False
        self.in_anchor_text = False
        
        self.title = None
        self.tag_level = 0
        self.block_tag_level = -1
        self.text_blocks = []
        self.label_stacks = []
        self.font_size_stack = []
        self.text_buffer = ''
        self.token_buffer = ''
        
        self.raise_on_failure = raise_on_failure
    
    def recycle(self) -> None:
        """
        Recycles this instance.
        """
        
        self.clear_text_buffer()
        self.in_body = 0
        self.in_anchor = 0
        self.in_ignorable_element = 0
        self.text_element_idx = 0
        self.last_start_tag = None
        self.last_end_tag = None
        self.last_event = None
        self.offset_blocks = 0
        self.current_contained_text_elements = set()
        self.flush = False
        self.in_anchor_text = False
        self.text_blocks = []
        
        # --------- added -------
        self.title = None
        self.tag_level = 0
        self.block_tag_level = -1
        self.label_stacks = []
        self.font_size_stack = []
    
    # ------------------------------- SAX Parser methods ----------------------------------------
    
    def end_document(self) -> None:
        self.flush_block()
    
    def start_document(self) -> None:
        pass
    
    def start_element(self, name: str, attrs: AttributesImpl) -> None:
        self.label_stacks.append([])
        
        tag_action = self.tagActions.get(name.strip().upper())
        if tag_action is not None:
            self.flush |= tag_action.start(self, name, attrs)
            if tag_action.changes_tag_level():
                self.tag_level += 1
        else:
            self.tag_level += 1
            self.flush = True
        self.last_event = self.EVENT_START_TAG
        self.last_start_tag = name
    
    def end_element(self, name: str) -> None:
        tag_action = self.tagActions.get(name.strip().upper())
        if tag_action is not None:
            self.flush |= tag_action.end(self, name)
            if tag_action.changes_tag_level():
                self.tag_level -= 1
        else:
            self.flush = True
            self.tag_level -= 1
        
        if self.flush:
            self.flush_block()
        self.last_event = self.EVENT_END_TAG
        self.last_end_tag = name
        try:
            self.label_stacks.pop()
        except IndexError:
            if self.raise_on_failure:
                raise
    
    def characters(self, content: str) -> None:
        self.text_element_idx += 1
        if self.flush:
            self.flush_block()
            self.flush = False
        if self.in_ignorable_element != 0:
            return
        
        if len(content) == 0:
            return
        
        stripped_content = content.strip()
        if len(stripped_content) == 0:
            self.add_whitespace_if_necessary()
            self.last_event = self.EVENT_WHITESPACE
            return
        
        start_whitespace = content[0].isspace()
        if start_whitespace:
            self.add_whitespace_if_necessary()
        
        if self.block_tag_level == -1:
            self.block_tag_level = self.tag_level
        self.text_buffer += stripped_content
        self.token_buffer += stripped_content
        
        end_whitespace = content[-1].isspace()
        if end_whitespace:
            self.add_whitespace_if_necessary()
        
        self.last_event = self.EVENT_CHARACTERS
        self.current_contained_text_elements.add(self.text_element_idx)
    
    def ignorable_whitespace(self) -> None:
        self.add_whitespace_if_necessary()
    
    # ------------------------------- utility methods ----------------------------------------
    
    def flush_block(self) -> None:
        if self.in_body == 0:
            if self.last_start_tag.lower() == "title":
                self.set_title(self.text_buffer.strip())
            self.clear_text_buffer()
            return
        if len(self.token_buffer.strip()) == 0:
            self.clear_text_buffer()
            return
        
        tokens = self.tokenize(self.token_buffer)
        num_words = 0
        num_linked_words = 0
        num_wrapped_lines = 0
        current_line_length = -1
        #  don't count the first space
        max_line_length = 80
        num_tokens = 0
        num_words_current_line = 0
        
        for token in tokens:
            if token == SpecialTokens.ANCHOR_TEXT_START:
                self.in_anchor_text = True
            elif token == SpecialTokens.ANCHOR_TEXT_END:
                self.in_anchor_text = False
            elif self.is_word(token):
                num_tokens += 1
                num_words += 1
                num_words_current_line += 1
                if self.in_anchor_text:
                    num_linked_words += 1
                current_line_length += len(token) + 1
                if current_line_length > max_line_length:
                    num_wrapped_lines += 1
                    current_line_length = len(token)
                    num_words_current_line = 1
            else:
                num_tokens += 1
        
        # if only special tokens (num_tokens excludes special tokens)
        if num_tokens == 0:
            self.clear_text_buffer()
            return
        
        if num_wrapped_lines == 0:
            num_words_in_wrapped_lines = num_words
            num_wrapped_lines = 1
        else:
            num_words_in_wrapped_lines = num_words - num_words_current_line
        
        tb = TextBlock(self.text_buffer.strip(), self.current_contained_text_elements, num_words, num_linked_words,
                       num_words_in_wrapped_lines, num_wrapped_lines, self.offset_blocks)
        self.current_contained_text_elements = set()
        self.offset_blocks += 1
        self.clear_text_buffer()
        tb.tag_level = self.block_tag_level
        self.add_text_block(tb)
        self.block_tag_level = -1
    
    def add_text_block(self, tb: TextBlock) -> None:
        for font_size in self.font_size_stack[::-1]:
            if font_size is not None:
                tb.add_label(f"font-{font_size}")
                break
        for label_stack in self.label_stacks:
            for labels in label_stack:
                labels.add_to(tb)
        self.text_blocks.append(tb)
    
    def is_word(self, token: str) -> bool:
        return self.PAT_VALID_WORD_CHARACTER.search(token) is not None
    
    def tokenize(self, text: str) -> list:
        return self.PAT_WORD.findall(text)
    
    def set_title(self, s: str) -> None:
        if s is None or len(s) == 0:
            return
        self.title = s
    
    def to_text_document(self) -> TextDocument:
        """
        Returns a TextDocument containing the extracted TextBlocks. NOTE: Only call this after parsing.
        
        :return: The TextDocument
        """
        
        #  just to be sure
        self.flush_block()
        return TextDocument(self.text_blocks, self.title)
    
    def add_whitespace_if_necessary(self) -> None:
        if len(self.text_buffer) == 0 or not self.text_buffer[-1].isspace():
            self.text_buffer += ' '
        if len(self.token_buffer) == 0 or not self.token_buffer[-1].isspace():
            self.token_buffer += ' '
    
    def clear_text_buffer(self) -> None:
        self.text_buffer = ''
        self.token_buffer = ''
    
    def add_token(self, token: str) -> None:
        self.add_whitespace_if_necessary()
        self.token_buffer += token
        self.add_whitespace_if_necessary()
    
    def add_label_action(self, la: LabelAction) -> None:
        if len(self.label_stacks) == 0:
            self.label_stacks.append([])
        self.label_stacks[-1].append(la)


class BoilerpipeHTMLParser(HTMLParser, BoilerpipeBaseParser):
    def __init__(self, raise_on_failure: bool = True) -> None:
        HTMLParser.__init__(self)
        BoilerpipeBaseParser.__init__(self, raise_on_failure=raise_on_failure)
    
    def feed(self, data: str) -> None:
        self.start_document()
        HTMLParser.feed(self, data)
        self.end_document()
    
    def handle_starttag(self, tag: str, attrs: AttributesImpl) -> None:
        self.start_element(tag, attrs)
    
    def handle_endtag(self, tag: str) -> None:
        self.end_element(tag)
    
    def handle_data(self, data: str) -> None:
        self.characters(data)


class BoilerpipeSAXContentHandler(ContentHandler, BoilerpipeBaseParser):
    def __init__(self, raise_on_failure: bool = True) -> None:
        ContentHandler.__init__(self)
        BoilerpipeBaseParser.__init__(self, raise_on_failure=raise_on_failure)



"""
This file is licensed under the terms of the Apache License, Version 2.0. See the LICENSE file in the root of this
repository for complete details.
"""

from boilerpy3 import document, extractors, filters, parser

__all__ = ['document', 'extractors', 'filters', 'parser']



from setuptools import setup

with open('README.md') as readme_file:
    long_description = readme_file.read()

setup(
        name='boilerpy3',
        version='1.0.4',
        python_requires='>=3.6.*',
        author='John Riebold',
        author_email='jmriebold@gmail.com',
        license='Apache 2.0',
        description='Python port of Boilerpipe, Boilerplate Removal and Fulltext Extraction from HTML pages',
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords=[
            'boilerpipe',
            'boilerpy',
            'html text extraction',
            'text extraction',
            'full text extraction'
        ],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Topic :: Utilities',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'
        ],
        url='https://github.com/jmriebold/BoilerPy3',
        packages=[
            'boilerpy3'
        ],
)
